/*
 *  ircd-hybrid: an advanced Internet Relay Chat Daemon(ircd).
 *  m_svsnick.c :
 *
 *  Copyright (C) 2003 Elias Shtift. ishtift@tagil.svrw.ru
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id $
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "irc_string.h"
#include "ircd.h"
#include "numeric.h"
#include "s_misc.h"
#include "s_conf.h"
#include "s_serv.h"
#include "s_user.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "hash.h"
#include "packet.h"
#include "channel.h"
#include "channel_mode.h"
#include "s_log.h"


static void m_svsnick(struct Client *client_p, struct Client *source_p,
                       int parc, char *parv[]);
static void m_svsmode(struct Client *client_p, struct Client *source_p,
                       int parc, char *parv[]);
static void m_svsjoin(struct Client *client_p, struct Client *source_p,
                       int parc, char *parv[]);
static void m_svspart(struct Client *client_p, struct Client *source_p,
                       int parc, char *parv[]);

int clean_nick_name(char *);
int clean_host_name(char *);

struct Message svsnick_msgtab = {
  "SVSNICK", 0, 0, 3, 0, MFLG_SLOW, 0,
  {m_ignore, m_ignore, m_svsnick, m_ignore, m_ignore}
};

struct Message svsmode_msgtab = {
  "SVSMODE", 0, 0, 3, 0, MFLG_SLOW, 0,
  {m_ignore, m_ignore, m_svsmode, m_ignore, m_ignore}
};

struct Message svsjoin_msgtab = {
  "SVSJOIN", 0, 0, 3, 0, MFLG_SLOW, 0,
  {m_ignore, m_ignore, m_svsjoin, m_ignore, m_ignore}
};

struct Message svspart_msgtab = {
  "SVSPART", 0, 0, 3, 0, MFLG_SLOW, 0,
  {m_ignore, m_ignore, m_svspart, m_ignore, m_ignore}
};

#ifndef STATIC_MODULES
void
_modinit(void)
{
  mod_add_cmd(&svsnick_msgtab);
  mod_add_cmd(&svsmode_msgtab);
  mod_add_cmd(&svsjoin_msgtab);
  mod_add_cmd(&svspart_msgtab);
}

void
_moddeinit(void)
{
  mod_del_cmd(&svsnick_msgtab);
  mod_del_cmd(&svsmode_msgtab);
  mod_del_cmd(&svsjoin_msgtab);
  mod_del_cmd(&svspart_msgtab);
}

const char *_version = "$Revision: 1.0 $";
#endif

/*
 * ms_svsnick()
 *
 *
 */

static void m_svsnick(struct Client *client_p, struct Client *source_p,
                      int parc, char *parv[])
{
  struct Client *target_p, *target2_p;

  if (!clean_nick_name(parv[2]))
    return;

  if (!find_u_conf((char *)source_p->name,
		   NULL, NULL))
    return;

  if (!(target_p = find_client(parv[1])))
    return;

  if (MyConnect(target_p))
  {
  
    if ((target2_p = find_client(parv[2])))
    {

    /* If(target_p == source_p) the client is changing nicks between
     * equivalent nicknames ie: [nick] -> {nick}
     */

      if (target_p == target2_p)
      {
        if (strcmp(target2_p->name, parv[2]))
	{
          change_local_nick(target_p, target_p, parv[2]);
	  return;
	}
	else
	{
	  return;
	}
      }

    /* if the client that has the nick isnt registered yet (nick but no
     * user) then drop the unregged client
     */

      if (IsUnknown(target2_p))
      {
        exit_client(NULL, target2_p, &me, "Overridden");
        change_local_nick(target_p, target_p, parv[2]);
        return;
      }
      else
      {
        return;
      }
    }
    else
    {
      if (!ServerInfo.hub && uplink && IsCapable(uplink, CAP_LL))
      {
        /* The uplink might know someone by this name already. */
        sendto_one(uplink, ":%s NBURST %s %s %s", me.name, parv[2],
                   parv[2], parv[1]);
        return;
      }
      else
      {
        change_local_nick(target_p, target_p, parv[2]);
	return;
      }
    }
  }
  else
  {
    sendto_server(client_p, source_p, NULL, CAP_SVS, NOCAPS, NOFLAGS,
                  ":%s SVSNICK %s %s", source_p->name, parv[1], parv[2]);
    return;
  }
}

/*
 * ms_svsmode()
 *
 *
 */

static void m_svsmode(struct Client *client_p, struct Client *source_p,
                      int parc, char *parv[])
{
  struct Client *target_p;
  char **p, *m;
  int what = 1, oldflags;

  if (!find_u_conf((char *)source_p->name,
		   NULL, NULL))
    return;

  if (!(target_p = find_client(parv[1])))
    return;

  if (!MyConnect(target_p))
  {
    sendto_server(client_p, source_p, NULL, CAP_SVS, NOCAPS, NOFLAGS,
                  ":%s SVSMODE %s %s", source_p->name, parv[1], parv[2]);
    return;
  }

  oldflags = target_p->umodes;

  for (p = &parv[2]; p && *p; p++)
  {
    for (m = *p; *m; m++)
    {
      switch (*m)
      {
        case '+':
          what = 1;
          break;
        case '-':
          what = 0;
          break;
        case 'R':
          if (what)
          {
	    target_p->umodes |= FLAGS_REGNICK;
          }
          else
          {
	    target_p->umodes &= ~FLAGS_REGNICK;
          }
	  break;
        case 'y':
          if (what)
          {
	    target_p->umodes |= FLAGS_SPY;
          }
          else
          {
	    target_p->umodes &= ~FLAGS_SPY;
          }
	  break;
      }
    }
  }

  send_umode_out(target_p, target_p, oldflags);
}

/*
 * ms_svsjoin()
 *
 *
 */

static void m_svsjoin(struct Client *client_p, struct Client *source_p,
                      int parc, char *parv[])
{
  struct Client *target_p;
  struct Channel *chptr;

  if (!find_u_conf((char *)source_p->name,
		   NULL, NULL))
    return;

  if (!(target_p = find_client(parv[1])))
    return;

  if (!MyConnect(target_p))
  {
    sendto_server(client_p, source_p, NULL, CAP_SVS, NOCAPS, NOFLAGS,
                  ":%s SVSJOIN %s %s", source_p->name, parv[1], parv[2]);
    return;
  }

  if (*parv[2] != '#')
    return;

  if ((chptr = hash_find_channel(parv[2])))
  {
    if (IsMember(target_p, chptr))
      return;

    add_user_to_channel(chptr, target_p, CHFL_PEON);

    if (chptr->chname[0] == '#')
      sendto_server(target_p, target_p, chptr, NOCAPS, NOCAPS, LL_ICLIENT,
                    ":%s SJOIN %lu %s + :%s",
		    me.name, (unsigned long)chptr->channelts,
		    chptr->chname, target_p->name);

    sendto_channel_local(ALL_MEMBERS, chptr, ":%s!%s@%s JOIN :%s",
                         target_p->name, target_p->username,
			 target_p->host, chptr->chname);

    if (chptr->topic)
    {
      sendto_one(target_p, form_str(RPL_TOPIC),
                 me.name, target_p->name,
		 chptr->chname, chptr->topic);

      sendto_one(target_p, form_str(RPL_TOPICWHOTIME),
                 me.name, source_p->name, chptr->chname,
		 chptr->topic_info, chptr->topic_time);
    }

    channel_member_names(target_p, chptr, chptr->chname, 1);
  }
  else
  {
    if (!check_channel_name(parv[2]))
      return;

    if (strlen(parv[2]) > CHANNELLEN)
      return;

    chptr = get_or_create_channel(target_p, parv[2], NULL);
    add_user_to_channel(chptr, target_p, CHFL_CHANOP);

    if (chptr->chname[0] == '#')
      sendto_server(target_p, target_p, chptr, NOCAPS, NOCAPS, LL_ICLIENT,
                    ":%s SJOIN %lu %s +nt :@%s",
		    me.name, (unsigned long)chptr->channelts,
		    chptr->chname, target_p->name);

    sendto_channel_local(ALL_MEMBERS, chptr, ":%s!%s@%s JOIN :%s",
                         target_p->name, target_p->username,
			 target_p->host, chptr->chname);

    chptr->mode.mode |= MODE_TOPICLIMIT | MODE_NOPRIVMSGS;

    sendto_channel_local(ALL_MEMBERS, chptr, ":%s MODE %s +nt",
                         me.name, chptr->chname);

    target_p->localClient->last_join_time = CurrentTime;
    channel_member_names(target_p, chptr, chptr->chname, 1);
  }
}

/*
 * ms_svspart()
 *
 *
 */

static void m_svspart(struct Client *client_p, struct Client *source_p,
                      int parc, char *parv[])
{
  struct Client *target_p;
  struct Channel *chptr;
  struct Membership *member;

  if (!find_u_conf((char *)source_p->name,
		   NULL, NULL))
    return;

  if (!(target_p = find_client(parv[1])))
    return;

  if (!MyConnect(target_p))
  {
    sendto_server(client_p, source_p, NULL, CAP_SVS, NOCAPS, NOFLAGS,
                  ":%s SVSPART %s %s", source_p->name, parv[1], parv[2]);
    return;
  }

  if (*parv[2] != '#')
    return;

  if((chptr = hash_find_channel(parv[2])) == NULL)
    return;

  if (!IsMember(target_p, chptr))
    return;
  
  if (chptr->chname[0] != '&')
    sendto_server(target_p, target_p, chptr, NOCAPS, NOCAPS, LL_ICLIENT,
		  ":%s PART %s :%s",
		  target_p->name, chptr->chname,
		  target_p->name);

  sendto_channel_local(ALL_MEMBERS, chptr, ":%s!%s@%s PART %s :%s",
                       target_p->name, target_p->username,
 	               target_p->host,chptr->chname,
		       target_p->name);
  remove_user_from_channel(chptr, target_p);

}

/* clean_nick_name()
 *
 * input	- nickname
 * output	- none
 * side effects - walks through the nickname, returning 0 if erroneous
 */
int clean_nick_name(char *nick)
{
  assert(nick);
  if (nick == NULL)
    return (0);

  /* nicks cant start with a digit or - or be 0 length */
  /* This closer duplicates behaviour of hybrid-6 */

  if (*nick == '-' || IsDigit(*nick) || *nick == '\0')
    return (0);

  for(; *nick; nick++)
  {
    if (!IsNickChar(*nick))
      return (0);
  }

  return (1);
}

/* clean_host_name()
 * input	- hostname
 * output	- none
 * side effects - walks through the hostname, returning 0 if erroneous
 */
int clean_host_name(char *host)
{
  assert(host);
  if (host == NULL)
    return 0;
  for(; *host; host++)
  {
    if (!IsHostChar(*host))
      return 0;
  }

  return 1;
}
