/************************************************************************
 *   IRC - Internet Relay Chat, src/match.c
 *   Copyright (C) 1990 Jarkko Oikarinen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: match.c,v 7.21 2002/08/28 17:00:14 androsyn Exp $
 *
 */
#include "stdinc.h"
#include "irc_string.h"
#include "config.h"     
#include "client.h"
#include "ircd.h"

/* match()
 * 
 *  Compare if a given string (name) matches the given
 *  mask (which can contain wild cards: '*' - match any
 *  number of chars, '?' - match any single character.
 *
 *      return  1, if match
 *              0, if no match
 *
 *  Originally by Douglas A Lewis (dalewis@acsu.buffalo.edu)
 */
#define MATCH_MAX_CALLS 512  /* ACK! This dies when it's less that this
                                and we have long lines to parse */
int match(const char *mask, const char *name)
{
  const unsigned char* m = (const unsigned char*)  mask;
  const unsigned char* n = (const unsigned char*)  name;
  const unsigned char* ma = (const unsigned char*) mask;
  const unsigned char* na = (const unsigned char*) name;
  int   wild  = 0;
  int   calls = 0;

  assert(mask != NULL);
  assert(name != NULL);

  if (!mask || !name)
    return 0;

  while (calls++ < MATCH_MAX_CALLS)
  {
    if (*m == '*')
    {
      /*
       * XXX - shouldn't need to spin here, the mask should have been
       * collapsed before match is called
       */
      while (*m == '*')
        m++;
      wild = 1;
      ma = m;
      na = n;
    }

    if (!*m)
    {
      if (!*n)
        return 1;
      for (m--; (m > (const unsigned char*) mask) && (*m == '?'); m--)
        ;
      if (*m == '*' && (m > (const unsigned char*) mask))
        return 1;
      if (!wild)
        return 0;
      m = ma;
      n = ++na;
    }
    else if (!*n)
    {
      /*
       * XXX - shouldn't need to spin here, the mask should have been
       * collapsed before match is called
       */
      while (*m == '*')
        m++;
      return (*m == 0);
    }
    if (ToIden(ToLower(*m)) != ToIden(ToLower(*n)) && *m != '?')
    {
      if (!wild)
        return 0;
      m = ma;
      n = ++na;
    }
    else
    {
      if (*m)
        m++;
      if (*n)
        n++;
    }
  }
  return 0;
}

/* match_esc()
 *
 * The match() function with support for escaping characters such
 * as '*' and '?'
 */
int match_esc(const char *mask, const char *name)
{
  const unsigned char* m = (const unsigned char*)  mask;
  const unsigned char* n = (const unsigned char*)  name;
  const unsigned char* ma = (const unsigned char*) mask;
  const unsigned char* na = (const unsigned char*) name;
  int   wild  = 0;
  int   calls = 0;
  int   quote = 0;

  assert(mask != NULL);
  assert(name != NULL);

  if (!mask || !name)
    return 0;

  while (calls++ < MATCH_MAX_CALLS)
  {
    if (quote)
      quote++;
    if (quote == 3)
      quote = 0;
    if (*m == '\\' && !quote)
    {
      m++;
      quote = 1;
      continue;
    }
    if (!quote && *m == '*')
    {
      /*
       * XXX - shouldn't need to spin here, the mask should have been
       * collapsed before match is called
       */
      while (*m == '*')
        m++;
      if (*m == '\\')
      {
	m++;
	/* This means it is an invalid mask -A1kmm. */
	if (!*m)
	  return 0;
	quote = 2;
      }
      wild = 1;
      ma = m;
      na = n;
    }

    if (!*m)
    {
      if (!*n)
        return 1;
      if (quote)
        return 0;
      for (m--; (m > (const unsigned char*) mask) && (*m == '?'); m--)
        ;
      if (*m == '*' && (m > (const unsigned char*) mask))
        return 1;
      if (!wild)
        return 0;
      m = ma;
      n = ++na;
    }
    else if (!*n)
    {
      /*
       * XXX - shouldn't need to spin here, the mask should have been
       * collapsed before match is called
       */
      if (quote)
        return 0;
      while (*m == '*')
        m++;
      return (*m == 0);
    }
    if (ToLower(*m) != ToLower(*n) && !(!quote && *m == '?')) {
      if (!wild)
        return 0;
      m = ma;
      n = ++na;
    }
    else
    {
      if (*m)
        m++;
      if (*n)
        n++;
    }
  }
  return 0;
}

static inline int 
comp_with_mask(void *addr, void *dest, u_int mask)
{
  if (memcmp(addr, dest, mask / 8) == 0) 
  {
    int n = mask / 8;
    int m = ((-1) << (8 - (mask % 8)));
    if (mask % 8 == 0 || 
       (((u_char *) addr)[n] & m) == (((u_char *) dest)[n] & m))  
      return (1);
  }
  return (0);
}


/* match_cidr()
 *
 * Input - mask, address
 * Ouput - 1 = Matched 0 = Did not match
 */

int
match_cidr(const char *s1, const char *s2)
{
  struct irc_inaddr ipaddr, maskaddr;
  char address[NICKLEN + USERLEN + HOSTLEN + 6], mask[NICKLEN + USERLEN + HOSTLEN + 6], *ipmask, *ip, *len;
  int cidrlen, aftype;
  strcpy(mask, s1);
  strcpy(address, s2);
  
  ipmask = strrchr(mask, '@');
  if(ipmask == NULL)
    return 0;
  
  *ipmask++ = '\0';
  
  ip = strrchr(address, '@');
  if(ip == NULL)
    return 0;
  *ip++ = '\0';
  
  len = strrchr(ipmask, '/');
  if(len == NULL)
    return 0;
  
  *len++ = '\0';
  
  cidrlen = atoi(len);
  if(cidrlen == 0) 
    return 0;
  
#ifdef IPV6
  if(strchr(ip, ':') && strchr(ipmask, ':'))
    aftype = AF_INET6;
  else
#endif
  if(!strchr(ip, ':') && !strchr(ipmask, ':'))
    aftype = AF_INET;
  else
    return 0;
  
  inetpton(aftype, ip, &ipaddr);
  inetpton(aftype, ipmask, &maskaddr);
  if(comp_with_mask(&IN_ADDR(ipaddr), &IN_ADDR(maskaddr), cidrlen) && match(mask, address))
    return 1;
  else
    return 0;
}

/* collapse()
 *
 * collapses a string containing multiple *'s.
 */
char *
collapse(char *pattern)
{
  char *p = pattern, *po = pattern;
  char c;
  int f = 0;

  if (p == NULL)
    return NULL;
 
  while ((c = *p++))
  {
    if (c == '*')
    {
      if (!(f & 1))
        *po++ = '*';
      f |= 1;
    }
    else
    {
      *po++ = c;
      f &= ~1;
    }
  }
  *po++ = 0;

  return pattern;
}

/* collapse_esc()
 *
 * The collapse() function with support for escaping characters
 */
char *collapse_esc(char *pattern)
{
 char *p = pattern, *po = pattern;
 char c;
 int f = 0;

 if (p == NULL)
   return NULL;
 
 while ((c = *p++))
 {
   if (!(f & 2) && c == '*')
   {
     if (!(f & 1))
       *po++ = '*';
     f |= 1;
   }
   else if (!(f & 2) && c == '\\')
   {
     *po++ = '\\';
     f |= 2;
   }
   else
   {
     *po++ = c;
     f &= ~3;
   }
 }
 *po++ = 0;
 return pattern;
}

/*
 * irccmp - case insensitive comparison of two 0 terminated strings.
 *
 *      returns  0, if s1 equal to s2
 *              <0, if s1 lexicographically less than s2
 *              >0, if s1 lexicographically greater than s2
 */
int
irccmp(const char *s1, const char *s2)
{
  const unsigned char *str1 = (const unsigned char *)s1;
  const unsigned char *str2 = (const unsigned char *)s2;

  assert(s1 != NULL);
  assert(s2 != NULL);
  
  while ((ToIden(ToUpper(*str1)) - ToIden(ToUpper(*str2))) == 0)
  {
    if (*str1 == '\0')
      return(0);
    str1++;
    str2++;
  }
  return(1);
}

int ircncmp(const char *s1, const char *s2, int n)
{
  const unsigned char* str1 = (const unsigned char*) s1;
  const unsigned char* str2 = (const unsigned char*) s2;
  int res;
  assert(s1 != NULL);
  assert(s2 != NULL);

  while ((res = ToIden(ToUpper(*str1)) - ToIden(ToUpper(*str2))) == 0)
  {
    str1++; str2++; n--;
    if (n == 0 || (*str1 == '\0' && *str2 == '\0'))
      return(0);
  }
  return(res);
}

const unsigned char ToLowerTab[] = {
  0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
  0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
  0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
  0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
  0x40, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F,
  0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
  0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F,
  0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F,
  0x90, 0x83, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x9A, 0x8B, 0x9C, 0x9D, 0x9E, 0x9F,
  0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
  0xA0, 0xA2, 0xA2, 0xDC, 0xA4, 0xB4, 0xA6, 0xA7, 0xB8, 0xA9, 0xBA, 0xAB, 0xAC, 0xAD, 0xAE, 0xBF,
  0xB0, 0xB1, 0xB3, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBE, 0xBE, 0xBF,
  0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF,
  0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF,
  0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF,
  0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF
};

const unsigned char ToUpperTab[] = {
  0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
  0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
  0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
  0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
  0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
  0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
  0x60, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
  0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F,
  0x80, 0x81, 0x82, 0x81, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,
  0x80, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x8A, 0x9B, 0x8C, 0x8D, 0x8E, 0x8F,
  0xA0, 0xA1, 0xA1, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF,
  0xB0, 0xB1, 0xB2, 0xB2, 0xA5, 0xB5, 0xB6, 0xB7, 0xA8, 0xB9, 0xAA, 0xBB, 0xA3, 0xBD, 0xBD, 0xAF,
  0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,
  0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,
  0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,
  0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF
};

/*
 *
 */

const unsigned char ToIdenTab[] = {
  /* 00 | ' ' alias ' ' */   0x00,
  /* 01 | ' ' alias ' ' */   0x01,
  /* 02 | ' ' alias ' ' */   0x02,
  /* 03 | ' ' alias ' ' */   0x03,
  /* 04 | ' ' alias ' ' */   0x04,
  /* 05 | ' ' alias ' ' */   0x05,
  /* 06 | ' ' alias ' ' */   0x06,
  /* 07 | ' ' alias ' ' */   0x07,
  /* 08 | ' ' alias ' ' */   0x08,
  /* 09 | ' ' alias ' ' */   0x09,
  /* 0A | ' ' alias ' ' */   0x0a,
  /* 0B | ' ' alias ' ' */   0x0b,
  /* 0C | ' ' alias ' ' */   0x0c,
  /* 0D | ' ' alias ' ' */   0x0d,
  /* 0E | ' ' alias ' ' */   0x0e,
  /* 0F | ' ' alias ' ' */   0x0f,
  /* 10 | ' ' alias ' ' */   0x10,
  /* 11 | ' ' alias ' ' */   0x11,
  /* 12 | ' ' alias ' ' */   0x12,
  /* 13 | ' ' alias ' ' */   0x13,
  /* 14 | ' ' alias ' ' */   0x14,
  /* 15 | ' ' alias ' ' */   0x15,
  /* 16 | ' ' alias ' ' */   0x16,
  /* 17 | ' ' alias ' ' */   0x17,
  /* 18 | ' ' alias ' ' */   0x18,
  /* 19 | ' ' alias ' ' */   0x19,
  /* 1A | ' ' alias ' ' */   0x1a,
  /* 1B | ' ' alias ' ' */   0x1b,
  /* 1C | ' ' alias ' ' */   0x1c,
  /* 1D | ' ' alias ' ' */   0x1d,
  /* 1E | ' ' alias ' ' */   0x1e,
  /* 1F | ' ' alias ' ' */   0x1f,
  /* 20 | ' ' alias ' ' */   0x20,
  /* 21 | '!' alias '!' */   0x21,
  /* 22 | '"' alias '"' */   0x22,
  /* 23 | '#' alias '#' */   0x23,
  /* 24 | '$' alias '$' */   0x24,
  /* 25 | '%' alias '%' */   0x25,
  /* 26 | '&' alias '&' */   0x26,
  /* 27 | ''' alias ''' */   0x27,
  /* 28 | '(' alias '(' */   0x28,
  /* 29 | ')' alias ')' */   0x29,
  /* 2A | '*' alias '*' */   0x2a,
  /* 2B | '+' alias '+' */   0x2b,
  /* 2C | ',' alias ',' */   0x2c,
  /* 2D | '-' alias '-' */   0x2d,
  /* 2E | '.' alias '.' */   0x2e,
  /* 2F | '/' alias '/' */   0x2f,
  /* 30 | '0' alias '0' */   0x30,
  /* 31 | '1' alias '1' */   0x31,
  /* 32 | '2' alias '2' */   0x32,
  /* 33 | '3' alias '3' */   0x33,
  /* 34 | '4' alias '4' */   0x34,
  /* 35 | '5' alias '5' */   0x35,
  /* 36 | '6' alias '6' */   0x36,
  /* 37 | '7' alias '7' */   0x37,
  /* 38 | '8' alias '8' */   0x38,
  /* 39 | '9' alias '9' */   0x39,
  /* 3A | ':' alias ':' */   0x3a,
  /* 3B | ';' alias ';' */   0x3b,
  /* 3C | '<' alias '<' */   0x3c,
  /* 3D | '=' alias '=' */   0x3d,
  /* 3E | '>' alias '>' */   0x3e,
  /* 3F | '?' alias '?' */   0x3f,
  /* 40 | '@' alias '@' */   0x40,
  /* 41 | 'A' alias 'A' */   0x41,
  /* 42 | 'B' alias 'B' */   0x42,
  /* 43 | 'C' alias 'C' */   0x43,
  /* 44 | 'D' alias 'D' */   0x44,
  /* 45 | 'E' alias 'E' */   0x45,
  /* 46 | 'F' alias 'F' */   0x46,
  /* 47 | 'G' alias 'G' */   0x47,
  /* 48 | 'H' alias 'H' */   0x48,
  /* 49 | 'I' alias 'I' */   0x49,
  /* 4A | 'J' alias 'J' */   0x4a,
  /* 4B | 'K' alias 'K' */   0x4b,
  /* 4C | 'L' alias 'L' */   0x4c,
  /* 4D | 'M' alias 'M' */   0x4d,
  /* 4E | 'N' alias 'N' */   0x4e,
  /* 4F | 'O' alias 'O' */   0x4f,
  /* 50 | 'P' alias 'P' */   0x50,
  /* 51 | 'Q' alias 'Q' */   0x51,
  /* 52 | 'R' alias 'R' */   0x52,
  /* 53 | 'S' alias 'S' */   0x53,
  /* 54 | 'T' alias 'T' */   0x54,
  /* 55 | 'U' alias 'U' */   0x55,
  /* 56 | 'V' alias 'V' */   0x56,
  /* 57 | 'W' alias 'W' */   0x57,
  /* 58 | 'X' alias 'X' */   0x58,
  /* 59 | 'Y' alias 'Y' */   0x59,
  /* 5A | 'Z' alias 'Z' */   0x5a,
  /* 5B | '[' alias '[' */   0x5b,
  /* 5C | '\' alias '\' */   0x5c,
  /* 5D | ']' alias ']' */   0x5d,
  /* 5E | '^' alias '^' */   0x5e,
  /* 5F | '_' alias '_' */   0x5f,
  /* 60 | '`' alias '`' */   0x60,
  /* 61 | 'a' alias 'A' */   0x41,
  /* 62 | 'b' alias 'B' */   0x42,
  /* 63 | 'c' alias 'C' */   0x43,
  /* 64 | 'd' alias 'd' */   0x64,
  /* 65 | 'e' alias 'E' */   0x45,
  /* 66 | 'f' alias 'f' */   0x66,
  /* 67 | 'g' alias 'g' */   0x67,
  /* 68 | 'h' alias 'H' */   0x48,
  /* 69 | 'i' alias 'i' */   0x69,
  /* 6A | 'j' alias 'j' */   0x6a,
  /* 6B | 'k' alias 'K' */   0x4b,
  /* 6C | 'l' alias 'l' */   0x6c,
  /* 6D | 'm' alias 'M' */   0x4d,
  /* 6E | 'n' alias 'N' */   0x4e,
  /* 6F | 'o' alias 'O' */   0x4f,
  /* 70 | 'p' alias 'P' */   0x50,
  /* 71 | 'q' alias 'q' */   0x71,
  /* 72 | 'r' alias 'r' */   0x72,
  /* 73 | 's' alias 's' */   0x73,
  /* 74 | 't' alias 'T' */   0x54,
  /* 75 | 'u' alias 'u' */   0x75,
  /* 76 | 'v' alias 'v' */   0x76,
  /* 77 | 'w' alias 'w' */   0x77,
  /* 78 | 'x' alias 'X' */   0x58,
  /* 79 | 'y' alias 'Y' */   0x59,
  /* 7A | 'z' alias 'z' */   0x7a,
  /* 7B | '{' alias '{' */   0x7b,
  /* 7C | '|' alias '|' */   0x7c,
  /* 7D | '}' alias '}' */   0x7d,
  /* 7E | '~' alias '~' */   0x7e,
  /* 7F | '' alias '' */   0x7f,
  /* 80 | '' alias '' */   0x80,
  /* 81 | '' alias '' */   0x81,
  /* 82 | '' alias '' */   0x82,
  /* 83 | '' alias '' */   0x83,
  /* 84 | '' alias '' */   0x84,
  /* 85 | '' alias '' */   0x85,
  /* 86 | '' alias '' */   0x86,
  /* 87 | '' alias '' */   0x87,
  /* 88 | '' alias '' */   0x88,
  /* 89 | '' alias '' */   0x89,
  /* 8A | '' alias '' */   0x8a,
  /* 8B | '' alias '' */   0x8b,
  /* 8C | '' alias '' */   0x8c,
  /* 8D | '' alias '' */   0x8d,
  /* 8E | '' alias '' */   0x4f,
  /* 8F | '' alias '' */   0x8f,
  /* 90 | '' alias '' */   0x90,
  /* 91 | '' alias '' */   0x91,
  /* 92 | '' alias '' */   0x92,
  /* 93 | '' alias '' */   0x93,
  /* 94 | '' alias '' */   0x94,
  /* 95 | '' alias '' */   0x95,
  /* 96 | '' alias '' */   0x96,
  /* 97 | '' alias '' */   0x97,
  /* 98 | '' alias '' */   0x98,
  /* 99 | '' alias '' */   0x99,
  /* 9A | '' alias '' */   0x9a,
  /* 9B | '' alias '' */   0x9b,
  /* 9C | '' alias '' */   0x9c,
  /* 9D | '' alias '' */   0x9d,
  /* 9E | '' alias '' */   0x9e,
  /* 9F | '' alias '' */   0x9f,
  /* A0 | '' alias '' */   0xa0,
  /* A1 | '' alias '' */   0xa1,
  /* A2 | '' alias '' */   0xa2,
  /* A3 | '' alias '' */   0xa3,
  /* A4 | '' alias '' */   0xa4,
  /* A5 | '' alias '' */   0xa5,
  /* A6 | '' alias '' */   0xa6,
  /* A7 | '' alias '' */   0xa7,
  /* A8 | '' alias '' */   0xa8,
  /* A9 | '' alias '' */   0xa9,
  /* AA | '' alias '' */   0xaa,
  /* AB | '' alias '' */   0xab,
  /* AC | '' alias '' */   0xac,
  /* AD | '' alias '' */   0xad,
  /* AE | '' alias '' */   0xae,
  /* AF | '' alias '' */   0xaf,
  /* B0 | '' alias '' */   0xb0,
  /* B1 | '' alias '' */   0xb1,
  /* B2 | '' alias '' */   0xb2,
  /* B3 | '' alias '' */   0xb3,
  /* B4 | '' alias '' */   0xb4,
  /* B5 | '' alias '' */   0xb5,
  /* B6 | '' alias '' */   0xb6,
  /* B7 | '' alias '' */   0xb7,
  /* B8 | '' alias '' */   0xb8,
  /* B9 | '' alias '' */   0xb9,
  /* BA | '' alias '' */   0xba,
  /* BB | '' alias '' */   0xbb,
  /* BC | '' alias '' */   0xbc,
  /* BD | '' alias '' */   0xbd,
  /* BE | '' alias '' */   0xbe,
  /* BF | '' alias '' */   0xbf,
  /* C0 | '' alias 'A' */   0x41,
  /* C1 | '' alias '' */   0xe1,
  /* C2 | '' alias 'B' */   0x42,
  /* C3 | '' alias '' */   0xe3,
  /* C4 | '' alias '' */   0xe4,
  /* C5 | '' alias 'E' */   0x45,
  /* C6 | '' alias '' */   0xe6,
  /* C7 | '' alias '' */   0x33,
  /* C8 | '' alias '' */   0xe8,
  /* C9 | '' alias '' */   0xc9,
  /* CA | '' alias 'K' */   0x4b,
  /* CB | '' alias '' */   0xeb,
  /* CC | '' alias 'M' */   0x4d,
  /* CD | '' alias 'H' */   0x48,
  /* CE | '' alias 'O' */   0x4f,
  /* CF | '' alias '' */   0x4e,
  /* D0 | '' alias 'P' */   0x50,
  /* D1 | '' alias 'C' */   0x43,
  /* D2 | '' alias 'T' */   0x54,
  /* D3 | '' alias 'Y' */   0x59,
  /* D4 | '' alias '' */   0xf4,
  /* D5 | '' alias 'X' */   0x58,
  /* D6 | '' alias '' */   0xf6,
  /* D7 | '' alias '' */   0xf7,
  /* D8 | '' alias '' */   0xf8,
  /* D9 | '' alias '' */   0xf9,
  /* DA | '' alias '' */   0xfa,
  /* DB | '' alias '' */   0xfb,
  /* DC | '' alias '' */   0xfc,
  /* DD | '' alias '' */   0xfd,
  /* DE | '' alias '' */   0xfe,
  /* DF | '' alias '' */   0xff,
  /* E0 | '' alias 'A' */   0x41,
  /* E1 | '' alias '' */   0xc1,
  /* E2 | '' alias 'B' */   0x42,
  /* E3 | '' alias '' */   0xc3,
  /* E4 | '' alias '' */   0xc4,
  /* E5 | '' alias 'E' */   0x45,
  /* E6 | '' alias '' */   0xc6,
  /* E7 | '' alias '' */   0x33,
  /* E8 | '' alias '' */   0xc8,
  /* E9 | '' alias '' */   0xc9,
  /* EA | '' alias 'K' */   0x4b,
  /* EB | '' alias '' */   0xcb,
  /* EC | '' alias 'M' */   0x4d,
  /* ED | '' alias 'H' */   0x48,
  /* EE | '' alias 'O' */   0x4f,
  /* EF | '' alias 'n' */   0x4e,
  /* F0 | '' alias 'P' */   0x50,
  /* F1 | '' alias 'C' */   0x43,
  /* F2 | '' alias 'T' */   0x54,
  /* F3 | '' alias 'Y' */   0x59,
  /* F4 | '' alias '' */   0xd4,
  /* F5 | '' alias 'X' */   0x58,
  /* F6 | '' alias '' */   0xd6,
  /* F7 | '' alias '' */   0xd7,
  /* F8 | '' alias '' */   0xd8,
  /* F9 | '' alias '' */   0xd9,
  /* FA | '' alias '' */   0xda,
  /* FB | '' alias '' */   0xdb,
  /* FC | '' alias '' */   0xdc,
  /* FD | '' alias '' */   0xdd,
  /* FE | '' alias '' */   0xde,
  /* FF | '' alias '' */   0xdf
};

/*
 * CharAttrs table
 *
 * NOTE: RFC 1459 sez: anything but a ^G, comma, or space is allowed
 * for channel names
 */
const unsigned int CharAttrs[] = {
/* 0  */     CNTRL_C,
/* 1  */     CNTRL_C|CHAN_C|NONEOS_C,
/* 2  */     CNTRL_C|CHAN_C|NONEOS_C,
/* 3  */     CNTRL_C|CHAN_C|NONEOS_C,
/* 4  */     CNTRL_C|CHAN_C|NONEOS_C,
/* 5  */     CNTRL_C|CHAN_C|NONEOS_C,
/* 6  */     CNTRL_C|CHAN_C|NONEOS_C,
/* 7 BEL */  CNTRL_C|NONEOS_C,
/* 8  \b */  CNTRL_C|CHAN_C|NONEOS_C,
/* 9  \t */  CNTRL_C|SPACE_C|CHAN_C|NONEOS_C,
/* 10 \n */  CNTRL_C|SPACE_C|CHAN_C|NONEOS_C|EOL_C,
/* 11 \v */  CNTRL_C|SPACE_C|CHAN_C|NONEOS_C,
/* 12 \f */  CNTRL_C|SPACE_C|CHAN_C|NONEOS_C,
/* 13 \r */  CNTRL_C|SPACE_C|CHAN_C|NONEOS_C|EOL_C,
/* 14 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 15 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 16 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 17 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 18 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 19 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 20 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 21 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 22 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 23 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 24 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 25 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 26 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 27 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 28 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 29 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 30 */     CNTRL_C|CHAN_C|NONEOS_C,
/* 31 */     CNTRL_C|CHAN_C|NONEOS_C,
/* SP */     PRINT_C|SPACE_C,
/* ! */      PRINT_C|KWILD_C|CHAN_C|NONEOS_C,
/* " */      PRINT_C|CHAN_C|NONEOS_C,
/* # */      PRINT_C|CHANPFX_C|CHAN_C|NONEOS_C,
/* $ */      PRINT_C|CHAN_C|NONEOS_C|USER_C,
/* % */      PRINT_C|CHAN_C|NONEOS_C,
/* & */      PRINT_C|CHANPFX_C|CHAN_C|NONEOS_C,
/* ' */      PRINT_C|CHAN_C|NONEOS_C,
/* ( */      PRINT_C|CHAN_C|NONEOS_C,
/* ) */      PRINT_C|CHAN_C|NONEOS_C,
/* * */      PRINT_C|KWILD_C|CHAN_C|NONEOS_C|SERV_C,
/* + */      PRINT_C|CHAN_C|NONEOS_C,
/* , */      PRINT_C|NONEOS_C,
/* - */      PRINT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* . */      PRINT_C|KWILD_C|CHAN_C|NONEOS_C|USER_C|HOST_C|SERV_C,
/* / */      PRINT_C|CHAN_C|NONEOS_C,
/* 0 */      PRINT_C|DIGIT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* 1 */      PRINT_C|DIGIT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* 2 */      PRINT_C|DIGIT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* 3 */      PRINT_C|DIGIT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* 4 */      PRINT_C|DIGIT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* 5 */      PRINT_C|DIGIT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* 6 */      PRINT_C|DIGIT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* 7 */      PRINT_C|DIGIT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* 8 */      PRINT_C|DIGIT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* 9 */      PRINT_C|DIGIT_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C,
/* : */      PRINT_C|CHAN_C|NONEOS_C|HOST_C,
/* ; */      PRINT_C|CHAN_C|NONEOS_C,
/* < */      PRINT_C|CHAN_C|NONEOS_C,
/* = */      PRINT_C|CHAN_C|NONEOS_C,
/* > */      PRINT_C|CHAN_C|NONEOS_C,
/* ? */      PRINT_C|KWILD_C|CHAN_C|NONEOS_C,
/* @ */      PRINT_C|KWILD_C|CHAN_C|NONEOS_C,
/* A */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* B */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* C */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* D */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* E */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* F */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* G */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* H */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* I */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* J */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* K */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* L */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* M */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* N */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* O */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* P */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* Q */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* R */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* S */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* T */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* U */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* V */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* W */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* X */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* Y */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* Z */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|UPPER_C|LATIN_C,
/* [ */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C,
/* \ */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C,
/* ] */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C,
/* ^ */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C,
/* _ */      PRINT_C|NICK_C|CHAN_C|NONEOS_C|USER_C,
/* ` */      PRINT_C|NICK_C|CHAN_C|NONEOS_C|USER_C,
/* a */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* b */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* c */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* d */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* e */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* f */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* g */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* h */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* i */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* j */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* k */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* l */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* m */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* n */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* o */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* p */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* q */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* r */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* s */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* t */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* u */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* v */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* w */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* x */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* y */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* z */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|HOST_C|LOWER_C|LATIN_C,
/* { */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C,
/* | */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C,
/* } */      PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C,
/* ~ */      PRINT_C|ALPHA_C|CHAN_C|NONEOS_C|USER_C,
/* del  */   CHAN_C|NONEOS_C,
/* 0x80 */   CHAN_C|NONEOS_C,
/* 0x81 */   CHAN_C|NONEOS_C,
/* 0x82 */   CHAN_C|NONEOS_C,
/* 0x83 */   CHAN_C|NONEOS_C,
/* 0x84 */   CHAN_C|NONEOS_C,
/* 0x85 */   CHAN_C|NONEOS_C,
/* 0x86 */   CHAN_C|NONEOS_C,
/* 0x87 */   CHAN_C|NONEOS_C,
/* 0x88 */   CHAN_C|NONEOS_C,
/* 0x89 */   CHAN_C|NONEOS_C,
/* 0x8A */   CHAN_C|NONEOS_C,
/* 0x8B */   CHAN_C|NONEOS_C,
/* 0x8C */   CHAN_C|NONEOS_C,
/* 0x8D */   CHAN_C|NONEOS_C,
/* 0x8E */   CHAN_C|NONEOS_C,
/* 0x8F */   CHAN_C|NONEOS_C,
/* 0x90 */   CHAN_C|NONEOS_C,
/* 0x91 */   CHAN_C|NONEOS_C,
/* 0x92 */   CHAN_C|NONEOS_C,
/* 0x93 */   CHAN_C|NONEOS_C,
/* 0x94 */   CHAN_C|NONEOS_C,
/* 0x95 */   CHAN_C|NONEOS_C,
/* 0x96 */   CHAN_C|NONEOS_C,
/* 0x97 */   CHAN_C|NONEOS_C,
/* 0x98 */   CHAN_C|NONEOS_C,
/* 0x99 */   CHAN_C|NONEOS_C,
/* 0x9A */   CHAN_C|NONEOS_C,
/* 0x9B */   CHAN_C|NONEOS_C,
/* 0x9C */   CHAN_C|NONEOS_C,
/* 0x9D */   CHAN_C|NONEOS_C,
/* 0x9E */   CHAN_C|NONEOS_C,
/* 0x9F */   CHAN_C|NONEOS_C,
/* 0xA0 */   CHAN_C|NONEOS_C,
/* 0xA1 */   CHAN_C|NONEOS_C,
/* 0xA2 */   CHAN_C|NONEOS_C,
/* 0xA3 */   CHAN_C|NONEOS_C,
/* 0xA4 */   CHAN_C|NONEOS_C,
/* 0xA5 */   CHAN_C|NONEOS_C,
/* 0xA6 */   CHAN_C|NONEOS_C,
/* 0xA7 */   CHAN_C|NONEOS_C,
/* 0xA8 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xA9 */   CHAN_C|NONEOS_C,
/* 0xAA */   CHAN_C|NONEOS_C,
/* 0xAB */   CHAN_C|NONEOS_C,
/* 0xAC */   CHAN_C|NONEOS_C,
/* 0xAD */   CHAN_C|NONEOS_C,
/* 0xAE */   CHAN_C|NONEOS_C,
/* 0xAF */   CHAN_C|NONEOS_C,
/* 0xB0 */   CHAN_C|NONEOS_C,
/* 0xB1 */   CHAN_C|NONEOS_C,
/* 0xB2 */   CHAN_C|NONEOS_C,
/* 0xB3 */   CHAN_C|NONEOS_C,
/* 0xB4 */   CHAN_C|NONEOS_C,
/* 0xB5 */   CHAN_C|NONEOS_C,
/* 0xB6 */   CHAN_C|NONEOS_C,
/* 0xB7 */   CHAN_C|NONEOS_C,
/* 0xB8 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xB9 */   CHAN_C|NONEOS_C,
/* 0xBA */   CHAN_C|NONEOS_C,
/* 0xBB */   CHAN_C|NONEOS_C,
/* 0xBC */   CHAN_C|NONEOS_C,
/* 0xBD */   CHAN_C|NONEOS_C,
/* 0xBE */   CHAN_C|NONEOS_C,
/* 0xBF */   CHAN_C|NONEOS_C,
/* 0xC0 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xC1 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xC2 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xC3 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xC4 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xC5 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xC6 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xC7 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xC8 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xC9 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xCA */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xCB */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xCC */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xCD */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xCE */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xCF */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xD0 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xD1 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xD2 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xD3 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xD4 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xD5 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xD6 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xD7 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xD8 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xD9 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xDA */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xDB */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xDC */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xDD */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xDE */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xDF */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|UPPER_C|NATIVE_C,
/* 0xE0 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xE1 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xE2 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xE3 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xE4 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xE5 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xE6 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xE7 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xE8 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xE9 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xEA */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xEB */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xEC */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xED */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xEE */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xEF */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xF0 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xF1 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xF2 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xF3 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xF4 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xF5 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xF6 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xF7 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xF8 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xF9 */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xFA */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xFB */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xFC */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xFD */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xFE */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
/* 0xFF */   PRINT_C|ALPHA_C|NICK_C|CHAN_C|NONEOS_C|USER_C|LOWER_C|NATIVE_C,
};
